#===============================================================
# Project: Slit Bridge 100 - Ledex Drag Sensors Interface.
#------------------------------------------------------
# By: Jim Abraham
# File: LD_sensors.py
# Language: python3, Thonny
# IP Address: 10.53.77.36/24 sub=255.255.255.192 gw=10.53.77.62
# Date: 12-20-22
# Revision: 1-30-23
#------------------------------------------------------
# Source file:
# /home/abe/Desktop/Python_Projects/Ledex_Drag_Sensor_Interface/LD_sensors.py
# Runtime file:/home/abe/LD_sensors.py
# Alarm file:  /home/abe/LD_sensor_data_files/sys.log
#------------------------------------------------------
# Purpose:
#------------------------------------------------------
# Eliminate PC and 2-16c cables going to Line 1 Wr Computer Room.
# Eliminate bridge buzer and light.
# Need to run a ethernet cable to bridge for read/write PLCs.
# Read CUT12 PLC register 40001 = bridge head enables, and load status.
# Write ML PLC registers for alarming operator.
#===============================================================
#!/usr/bin/python3
#------------------------------------------------------
# Head_Drag_ALarms/No_Score_ALarms = variable bits set according to sensor in alarm.
# sb100_Head_Enables/Load = CUT12 PLC register 40001. Read modbus TCP/IP.
#LD_Sensor_h[1-4] = 0 when in alarm.
#------------------------------------------------------
import time
import os
from datetime import datetime
import RPi.GPIO as GPIO
GPIO.setwarnings(False)
GPIO.setmode(GPIO.BOARD)
#---------------------------------------
# PLC IP Address Assignments
#---------------------------------------
from pyModbusTCP.client import ModbusClient
Cut12PLC='10.53.77.51'
Cut34PLC='10.53.77.52'
Cut_PLC=Cut12PLC        #change to Cut34_PLC for SB300/400
MLPLC='10.53.10.168'    #remove cut12plc when done testing

#---------------------------------------
#PLC Register Assignments:
#---------------------------------------
# Enables in cutting system plcs.
cut12_plc_reg1=0 #40001 #sb100
cut34_plc_reg1=0 #40001 #sb300
cut34_plc_reg2=1 #40002 #sb400
cut_plc_reg1 = cut12_plc_reg1 #change to cut34_plc_reg1/2 for SB300/400

#---------------------------------------
#no score/head drag alarm registers in mainline plc.
#NOTE: Subtract 1 from register number.
#---------------------------------------
# SB100 = 412355/412356/412661
# SB300 = 41xxx/41yyy/41zzz
# SB400 = 41xxx/41yyy/41zzz
ml_plc_reg1=12354 #412355    #change,no score alarms
ml_plc_reg2=12355 #412356    #change,head drag alarms
ml_plc_reg3=12660 #412661    #change,WDT=1 (bit 16 in plc) 10sec update. 

# note: SB100 use Cut12PLC, SB300/400 use cut34PLC for c_cut host
c_cut = ModbusClient(host=Cut_PLC, port=502, auto_open=True, debug=False)
c_ml = ModbusClient(host=MLPLC, port=502, auto_open=True, debug=False)
#---------------------------------------
# variables
#---------------------------------------
bn= [0,1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,0] #0-14 head bit values
head_en= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]          # 14 heads enables,b15 load
sensor_stat= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]      # 14 heads (up/down)
BL0= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]              # 14 heads (head drag flag0)
BL1= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]              # 14 heads (head drag flag1)
BL2= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]              # 14 heads (no score flag2)
ledex_up_cnt= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]     # 14 heads (up/down counter)
ledex_down_cnt= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]   # 14 heads (up/down counter)
head_went_up= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]     # 14 heads (up flag)
drag_sensor_OK= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]   # 14 sensor status 0,1,2
Head_Drag_ALarms= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0] # 14 alarm=1
No_Score_ALarms= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]  # 14 alarm=1

Head_Drag_Alarm_PLC_Out=0  # alarm register bits
No_Score_Alarm_PLC_Out=0   # alarm register bits
Bridge_Loaded=0            # 40001 bit15 1=load, 0=unload
scan_delay=4               # sensor same for 4 cycles
wdt_delay=0.5              # flash WDT led
cycle_up_down_ok_cnt=4     #cycle 4 times = ok
ds=0
ds2=0
cut12plc_no_com=0
filemade=0
data_in_file=0
wdt_10sec=0
log_filename="/home/abe/LD_sensor_data_files/sys.log"
#---------------------------------------
# GP input Assignments:
#---------------------------------------
LD_Sensor_h= [0,10,3,5,7,29,31,26,24,21,19,23,32,33,8,0]   #0-14
#LD_Sensor_h[0]=0
#LD_Sensor_h[1]=10  #GPIO15
#LD_Sensor_h[2]=3   #GPIO02
#LD_Sensor_h[3]=5   #GPIO03
#LD_Sensor_h[4]=7   #GPIO04
#LD_Sensor_h[5]=29  #GPIO05
#LD_Sensor_h[6]=31  #GPIO06
#LD_Sensor_h[7]=26  #GPIO07
#LD_Sensor_h[8]=24  #GPIO08
#LD_Sensor_h[9]=21  #GPIO09
#LD_Sensor_h[10]=19 #GPIO10
#LD_Sensor_h[11]=23 #GPIO11
#LD_Sensor_h[12]=32 #GPIO12
#LD_Sensor_h[13]=33 #GPIO13
#LD_Sensor_h[14]=8  #GPIO14

#---------------------------------------
# GP output Assignments:
#---------------------------------------
RX_data= 11         #GPIO17
TX_data= 12         #GPIO18
LD_Sensor_Alarm=15  #GPIO22
WDT=16              #GPIO23
load_signal_out= 22 #GPIO25

#==============================================================
# SETUP:
#==============================================================

#---------------------------------------
# Define GP inputs
#---------------------------------------
for x in range(14): #0-14
    GPIO.setup(LD_Sensor_h[x+1], GPIO.IN, pull_up_down=GPIO.PUD_UP) #0=alarm,1=ok
    
#---------------------------------------
# Define GP outputs
#---------------------------------------
GPIO.setup(RX_data, GPIO.OUT)         #GPIO17
GPIO.setup(TX_data, GPIO.OUT)         #GPIO18
GPIO.setup(LD_Sensor_Alarm, GPIO.OUT) #GPIO22
GPIO.setup(WDT, GPIO.OUT)             #GPIO23
GPIO.setup(load_signal_out, GPIO.OUT) #GPIO25

#==============================================================
# Routines
#==============================================================

#---------------------------------------------------------
# Get Bridge Head 1-14 Enables, read register bits.
#---------------------------------------------------------
#sb100 read 40001 CUT12PLC
#sb300 read 40001 CUT34PLC
#sb400 read 40002 CUT34PLC
#---------------------------------------------------------
def read_CUT_PLC():
#---------------------------------------------------------
    global Bridge_Loaded
    global cut12plc_no_com
    GPIO.output(RX_data,1)
    regs_l = c_cut.read_holding_registers(cut_plc_reg1, 1) # (40001, 1 reg)
    #if success display registers
    if regs_l:
            #check bit value in register
        for bL in range(15):           #0-15 (16 bits)
            if (regs_l[0] & (1<<bL)):  #bL=bit location, check if bit=1 or 0
                head_en[bL+1]=1	       #head enabled  1-16 = head 101-114 enable status
            else:
                head_en[bL+1]=0	       #head not selected
            
            if cut12plc_no_com==1:
                cut12plc_no_com=0
                Write_to_file(f' Reading CUT12 PLC registers again!')
    else:
        if cut12plc_no_com==0:          
            Write_to_file(f' unable to read CUT12 PLC registers')
            cut12plc_no_com=1
        for bL in range(15): head_en[bL+1]=0
        for bL in range(4):
            GPIO.output(load_signal_out,1)
            time.sleep(0.1)
            GPIO.output(load_signal_out,0)
            time.sleep(0.1)
        
    Bridge_Loaded = head_en[15]  #get bridge load status          
    if Bridge_Loaded>0:
        GPIO.output(load_signal_out,1)
    else:
        GPIO.output(load_signal_out,0)
        
    GPIO.output(RX_data,0)

#---------------------------------------------------------
def update_Alarming():
#---------------------------------------------------------
    GPIO.output(TX_data,1)
    c_ml.write_single_register(ml_plc_reg1,No_Score_Alarm_PLC_Out)  #write to 412355
    c_ml.write_single_register(ml_plc_reg2,Head_Drag_Alarm_PLC_Out) #write to 412356  
    GPIO.output(TX_data,0)
    if No_Score_Alarm_PLC_Out>0 or Head_Drag_Alarm_PLC_Out>0:
        GPIO.output(LD_Sensor_Alarm,1)
    else:
        GPIO.output(LD_Sensor_Alarm,0)
        
#---------------------------------------------------------
def get_Sensor_Stat():
#---------------------------------------------------------
    for s in range(14): #0-14
        time.sleep(0.05)
        if GPIO.input(LD_Sensor_h[s+1]) == 1:
            sensor_stat[s+1]=1   #head down
        else:
            sensor_stat[s+1]=0   #head up

#---------------------------------------------------------
def check_if_heads_In_Alarm():
#---------------------------------------------------------
# Check sensor for proper head operation
#---------------------------------------------------------
# ledex_down_cnt =  count 4 up cycles to determine sensor working again.
# ledex_up_cnt =    count 4 down cycles to determine sensor working again.
# drag_sensor_OK =  0=ok, 1=no score, 2=head drag
# Bridge_Loaded =   bridge lower signal. cut12 plc
# head_en[] =       head enabled.        cut12 plc
# sensor_stat[] =   head up or down status
# head_went_down =  head went down properly
# head_went_up =    head went up properly
#---------------------------------------------------------
# check sensor operation: set alarm if sensor not working properly
# Bl	EN	U	D
# 0		0	0	1 #1
# 0		1	0	1 #1
# 1		0	0	1 #2
# 1		1	1	0 #3
#---------------------------------------------------------
# no alarm, set alarm if sensor not working.
#---------------------------------------------------------
    global ds
    for dsc in range(14): #0-14
        ds=dsc+1
        if drag_sensor_OK[ds]==0:    # no alarm
            #Bridge not loaded and head not enabled = head up ok
            #add delay counts to make sure load and sensor have time to make
            if Bridge_Loaded==0:
                if sensor_stat[ds]==1:#1 head down BAD
                    BL0[ds]+=1
                    BL1[ds]=0
                    BL2[ds]=0
                    if BL0[ds]>=scan_delay: #scan delay
                        BL0[ds]=0
                        drag_sensor_OK[ds]=2  #head drag alarm
                        Head_Drag_ALarms[ds]=bn[ds]
                        ds100=ds+100
                        Write_to_file(f' head#{ds100} Head Drag')
                        
                else:
                    BL0[ds]=0
                    BL1[ds]=0
                    BL2[ds]=0
            
            #Bridge loaded and head enabled = head up ok            
            elif Bridge_Loaded==1 and head_en[ds]==0:
                if sensor_stat[ds]==1: #2 head down BAD   
                    BL1[ds]+=1
                    BL0[ds]=0
                    if BL1[ds]>=scan_delay: #delay
                        BL1[ds]=0
                        drag_sensor_OK[ds]=3  #head drag alarm
                        Head_Drag_ALarms[ds]=bn[ds]
                        ds100=ds+100
                        Write_to_file(f' head#{ds100} Head Drag')
                else:
                    BL0[ds]=0
                    BL1[ds]=0
                    BL2[ds]=0
                    
            #Bridge loaded and head not enabled = head down ok
            elif Bridge_Loaded==1 and head_en[ds]==1:
                if sensor_stat[ds]==0: #3 head up BAD
                    BL2[ds]+=1
                    BL0[ds]=0
                    if BL2[ds]>=scan_delay: #delay
                        BL2[ds]=0
                        drag_sensor_OK[ds]=1 # no score alarm
                        No_Score_ALarms[ds]=bn[ds]
                        ds100=ds+100
                        Write_to_file(f' head#{ds100} No Score')
                else:
                    BL0[ds]=0
                    BL1[ds]=0
                    BL2[ds]=0

#---------------------------------------------------------
def check_if_heads_OK():
#---------------------------------------------------------
# Check if Head Drag/No Score alarm
# if in alarm, now try to clear alarm after sensor working
# again successfully 3 times.
# Bl	EN	U	D
# 0		0	1	0 #1 ok up
# 0		1	1	0 #1 ok up
# 1		0	1	0 #2 ok up
# 1		1	0	1 #3 ok down
#---------------------------------------------------------
# no score or head drag in alarm
#---------------------------------------------------------
    global Head_Drag_Alarm_PLC_Out
    global No_Score_Alarm_PLC_Out
    global ds2

    Head_Drag_Alarm_PLC_Out=0
    No_Score_Alarm_PLC_Out=0
    for dsc2 in range(14): #0-14
        ds2=dsc2+1
        if drag_sensor_OK[ds2]>0:  # an alarm occurred
            #Bridge not loaded and head not enabled = should be up
            if Bridge_Loaded==0 and head_went_up[ds2]==0:
                if sensor_stat[ds2]==0: #1 head up GOOD
                    ledex_up_cnt[ds2]+=1
                    head_went_up[ds2]=1
                else:       
                    ledex_up_cnt[ds2]=0 # sensor still bad
                    ledex_down_cnt[ds2]=0
                    head_went_up[ds2]=0
                    
            elif Bridge_Loaded==1 and head_went_up[ds2]==1:
                #Bridge loaded and head enabled = should be up
                if head_en[ds2]==0:
                    if sensor_stat[ds2]==0: #2a head up GOOD
                        ledex_up_cnt[ds2]+=1
                        head_went_up[ds2]=0
                    else:       
                        ledex_up_cnt[ds2]=0 # sensor still bad
                        ledex_down_cnt[ds2]=0
                        head_went_up[ds2]=0
                        
                #Bridge loaded and head not enabled = should be down       
                elif head_en[ds2]==1:
                    if sensor_stat[ds2]==1: #2b head down GOOD
                        ledex_down_cnt[ds2]+=1
                        head_went_up[ds2]=0
                    else:       
                        ledex_up_cnt[ds2]=0 # sensor still bad
                        ledex_down_cnt[ds2]=0
                        head_went_up[ds2]=0
               
    #---------------------------------------------------------
    # Reset alarm sensor working properly
    #---------------------------------------------------------
                if ledex_down_cnt[ds2]>=cycle_up_down_ok_cnt or ledex_up_cnt[ds2]>=cycle_up_down_ok_cnt:
                    if drag_sensor_OK[ds2]==2 or drag_sensor_OK[ds2]==3:
                        ds100=ds+100
                        Write_to_file(f' head#{ds100} Head Drag-OK!')
                    
                    if drag_sensor_OK[ds2]==1:
                        ds100=ds+100
                        Write_to_file(f' head#{ds100} No Score-OK!')
                    
                    drag_sensor_OK[ds2]=0
                    ledex_down_cnt[ds2]=0
                    ledex_up_cnt[ds2]=0
                    head_went_up[ds2]=0
                    Head_Drag_ALarms[ds2]=0
                    No_Score_ALarms[ds2]=0
                
        Head_Drag_Alarm_PLC_Out = Head_Drag_Alarm_PLC_Out + Head_Drag_ALarms[ds2]
        No_Score_Alarm_PLC_Out = No_Score_Alarm_PLC_Out + No_Score_ALarms[ds2]

#---------------------------------------------------------
def WDT_R():
#---------------------------------------------------------
    global wdt_10sec
    GPIO.output(WDT,1)
    time.sleep(wdt_delay)
    GPIO.output(WDT,0)
    time.sleep(wdt_delay)
    wdt_10sec+=1
    if wdt_10sec >=10:
        c_ml.write_single_register(ml_plc_reg3,1) # write a 1 to register for WDT alarming.   
        wdt_10sec=0
#---------------------------------------------------------
def Write_to_file(file_msg):
#---------------------------------------------------------
#Write to data File.
    data_in_file=1
    now = datetime.now()      #Get timestamp.
    f = open(log_filename, "a")
    f.write (str(now.strftime("%m/%d/%Y %H:%M:%S")))
    f.write (file_msg+"\n")  #timestamp then message.
    f.close()

#---------------------------------------------------------
def Rename_filename():
#---------------------------------------------------------
    global filemade
    now = datetime.now()      		#Get timestamp.
    now_hr=now.strftime("%H %p")	# 12 AM rename file
    if now_hr=="00 AM" and filemade==0 and data_in_file==1:
        new_filename="/home/abe/"+now.strftime("%m-%d-%Y_%H:%M:%S")+".log"
        filemade=1
        data_in_file=0
        os.rename(log_filename,new_filename)
        f = open(log_filename, "w") # only need to test renaming file.
        f.close()					# only need to test renaming file.
    if filemade==1 and now_hr!="00 AM":
        filemade=0
        
#==============================================================
# MAIN LOOP:
#==============================================================
Write_to_file(" :Reboot")
while True:
    read_CUT_PLC()      #Get Head Enables (enabled=1, not selected=0), and load signal
    get_Sensor_Stat()   #Get Ledex Drag Sensor (up=0, down=1)
    check_if_heads_In_Alarm() #Check if no score or head drag alarm present
    check_if_heads_OK() #Alarm clear?
    update_Alarming()   #Update Sensor alarming. (write to ml plc)
    WDT_R()
    Rename_filename()   # rename file at every midnight.
#==============================================================
    
#---------------------------------------
GPIO.cleanup()
#---------------------------------------
